@@ -14,6 +14,6 @@ for (var i = 0; i < 1000; i++) {
   })
 }
 
-y.argv
+y.parse()
 
 console.log('never get here')
