@@ -15,11 +15,19 @@ const noop = () => {}
 const implicationsFailedPattern = new RegExp(english['Implications failed:'])
 
 describe('yargs dsl tests', () => {
+  const oldProcess = { versions: {} }
+
   beforeEach(() => {
+    oldProcess.argv = process.argv
+    oldProcess.defaultApp = process.defaultApp
+    oldProcess.versions.electron = process.versions.electron
     yargs = require('../')
   })
 
   afterEach(() => {
+    process.argv = oldProcess.argv
+    process.defaultApp = oldProcess.defaultApp
+    process.versions.electron = oldProcess.versions.electron
     delete require.cache[require.resolve('../')]
   })
 
@@ -28,10 +36,38 @@ describe('yargs dsl tests', () => {
     process.env._ = '/usr/local/bin/ndm'
     process.execPath = '/usr/local/bin/ndm'
     const argv = yargs([]).parse()
-    argv['$0'].should.equal('ndm')
+    argv.$0.should.equal('ndm')
     yargs.$0.should.equal('ndm')
   })
 
+  it('should not remove the 1st argument of bundled electron apps', () => {
+    delete require.cache[require.resolve('../')]
+    process.argv = ['/usr/local/bin/app', '-f', 'toto', 'tutu']
+    process.versions.electron = '10.0.0-nightly.20200211'
+    yargs = require('../')
+    const argv = yargs.parse()
+    argv.should.have.property('f')
+    argv.f.should.equal('toto')
+    argv._.should.deep.equal(['tutu'])
+  })
+
+  it('should remove the 1st argument of unbundled electron apps', () => {
+    delete require.cache[require.resolve('../')]
+    process.argv = ['/usr/local/bin/electron', 'app.js', '-f', 'toto', 'tutu']
+    process.versions.electron = '10.0.0-nightly.20200211'
+    // Same syntax as in electron
+    Object.defineProperty(process, 'defaultApp', {
+      configurable: false,
+      enumerable: true,
+      value: true
+    })
+    yargs = require('../')
+    const argv = yargs.parse()
+    argv.should.have.property('f')
+    argv.f.should.equal('toto')
+    argv._.should.deep.equal(['tutu'])
+  })
+
   it('accepts an object for aliases', () => {
     const argv = yargs([])
       .alias({
@@ -286,6 +322,7 @@ describe('yargs dsl tests', () => {
         hiddenOptions: [],
         demandedCommands: {},
         demandedOptions: {},
+        deprecatedOptions: {},
         local: [
           '_',
           'foo',
@@ -543,11 +580,11 @@ describe('yargs dsl tests', () => {
 
   describe('terminalWidth', () => {
     it('returns the maximum width of the terminal', function () {
-      if (!process.stdout.isTTY) {
-        return this.skip()
+      if (process.stdout.isTTY) {
+        yargs.terminalWidth().should.be.gte(0)
+      } else {
+        expect(yargs.terminalWidth()).to.equal(null)
       }
-
-      yargs.terminalWidth().should.be.gte(0)
     })
   })
 
@@ -850,7 +887,7 @@ describe('yargs dsl tests', () => {
     })
 
     it('can be called multiple times with the same behavior', () => {
-      let counter = { foobar: 0 }
+      const counter = { foobar: 0 }
       yargs(['test', 'foobar'])
         .command(
           'test <name>',
@@ -2391,4 +2428,76 @@ describe('yargs dsl tests', () => {
       }, 5)
     })
   })
+
+  // See: https://github.com/yargs/yargs/issues/1098
+  it('should allow array and requires arg to be used in conjunction', () => {
+    const argv = yargs(['-i', 'item1', 'item2', 'item3'])
+      .option('i', {
+        alias: 'items',
+        type: 'array',
+        requiresArg: true
+      })
+      .argv
+    argv.items.should.eql(['item1', 'item2', 'item3'])
+    argv.i.should.eql(['item1', 'item2', 'item3'])
+  })
+
+  // See: https://github.com/yargs/yargs/issues/1570
+  describe('"nargs" with "array"', () => {
+    it('should not consume more than nargs items', () => {
+      const argv = yargs(['-i', 'item1', 'item2', '-i', 'item3', 'item4'])
+        .option('i', {
+          alias: 'items',
+          type: 'array',
+          nargs: 1
+        })
+        .argv
+      argv.items.should.eql(['item1', 'item3'])
+      argv.i.should.eql(['item1', 'item3'])
+      argv._.should.eql(['item2', 'item4'])
+    })
+
+    it('should apply nargs with higher precedence than requiresArg: true', () => {
+      const argv = yargs(['-i', 'item1', 'item2', '-i', 'item3', 'item4'])
+        .option('i', {
+          alias: 'items',
+          type: 'array',
+          nargs: 1,
+          requiresArg: true
+        })
+        .argv
+      argv.items.should.eql(['item1', 'item3'])
+      argv.i.should.eql(['item1', 'item3'])
+      argv._.should.eql(['item2', 'item4'])
+    })
+
+    // TODO: make this work with aliases, using a check similar to
+    // checkAllAliases() in yargs-parser.
+    it('should apply nargs with higher precedence than requiresArg()', () => {
+      const argv = yargs(['-i', 'item1', 'item2', '-i', 'item3', 'item4'])
+        .option('items', {
+          alias: 'i',
+          type: 'array',
+          nargs: 1
+        })
+        .requiresArg(['items'])
+        .argv
+      argv.items.should.eql(['item1', 'item3'])
+      argv.i.should.eql(['item1', 'item3'])
+      argv._.should.eql(['item2', 'item4'])
+    })
+
+    it('should raise error if not enough values follow nargs key', (done) => {
+      yargs
+        .option('i', {
+          alias: 'items',
+          type: 'array',
+          nargs: 1
+        })
+        .parse(['-i'], (err) => {
+          err.message.should.match(/Not enough arguments following: i/)
+          return done()
+        })
+    })
+  })
 })
