@@ -1,4 +1,9 @@
 'use strict'
+
+// an async function fails early in Node.js versions prior to 8.
+async function requiresNode8OrGreater () {}
+requiresNode8OrGreater()
+
 const argsert = require('./lib/argsert')
 const fs = require('fs')
 const Command = require('./lib/command')
@@ -11,7 +16,9 @@ const Y18n = require('y18n')
 const objFilter = require('./lib/obj-filter')
 const setBlocking = require('set-blocking')
 const applyExtends = require('./lib/apply-extends')
+const { globalMiddlewareFactory } = require('./lib/middleware')
 const YError = require('./lib/yerror')
+const processArgv = require('./lib/process-argv')
 
 exports = module.exports = Yargs
 function Yargs (processArgs, cwd, parentRequire) {
@@ -21,30 +28,44 @@ function Yargs (processArgs, cwd, parentRequire) {
   let command = null
   let completion = null
   let groups = {}
+  const globalMiddleware = []
   let output = ''
-  let preservedGroups = {}
+  const preservedGroups = {}
   let usage = null
   let validation = null
+  let handlerFinishCommand = null
 
   const y18n = Y18n({
     directory: path.resolve(__dirname, './locales'),
     updateFiles: false
   })
 
+  self.middleware = globalMiddlewareFactory(globalMiddleware, self)
+
   if (!cwd) cwd = process.cwd()
 
-  self.$0 = process.argv
-    .slice(0, 2)
+  self.scriptName = function (scriptName) {
+    self.customScriptName = true
+    self.$0 = scriptName
+    return self
+  }
+
+  // ignore the node bin, specify this in your
+  // bin file with #!/usr/bin/env node
+  if (/\b(node|iojs|electron)(\.exe)?$/.test(process.argv[0])) {
+    self.$0 = process.argv.slice(1, 2)
+  } else {
+    self.$0 = process.argv.slice(0, 1)
+  }
+
+  self.$0 = self.$0
     .map((x, i) => {
-      // ignore the node bin, specify this in your
-      // bin file with #!/usr/bin/env node
-      if (i === 0 && /\b(node|iojs)(\.exe)?$/.test(x)) return
       const b = rebase(cwd, x)
       return x.match(/^(\/|([a-zA-Z]:)?\\)/) && b.length < x.length ? b : x
     })
     .join(' ').trim()
 
-  if (process.env._ !== undefined && process.argv[1] === process.env._) {
+  if (process.env._ !== undefined && processArgv.getProcessArgvBin() === process.env._) {
     self.$0 = process.env._.replace(
       `${path.dirname(process.execPath)}/`, ''
     )
@@ -80,25 +101,30 @@ function Yargs (processArgs, cwd, parentRequire) {
       })
     })
 
-    // preserve all groups not set to local.
-    preservedGroups = Object.keys(groups).reduce((acc, groupName) => {
-      const keys = groups[groupName].filter(key => !(key in localLookup))
-      if (keys.length > 0) {
-        acc[groupName] = keys
-      }
-      return acc
-    }, {})
+    // add all groups not set to local to preserved groups
+    Object.assign(
+      preservedGroups,
+      Object.keys(groups).reduce((acc, groupName) => {
+        const keys = groups[groupName].filter(key => !(key in localLookup))
+        if (keys.length > 0) {
+          acc[groupName] = keys
+        }
+        return acc
+      }, {})
+    )
     // groups can now be reset
     groups = {}
 
     const arrayOptions = [
-      'array', 'boolean', 'string', 'requiresArg', 'skipValidation',
-      'count', 'normalize', 'number'
+      'array', 'boolean', 'string', 'skipValidation',
+      'count', 'normalize', 'number',
+      'hiddenOptions'
     ]
 
     const objectOptions = [
       'narg', 'key', 'alias', 'default', 'defaultDescription',
-      'config', 'choices', 'demandedOptions', 'demandedCommands', 'coerce'
+      'config', 'choices', 'demandedOptions', 'demandedCommands', 'coerce',
+      'deprecatedOptions'
     ]
 
     arrayOptions.forEach((k) => {
@@ -116,7 +142,7 @@ function Yargs (processArgs, cwd, parentRequire) {
     // instances of all our helpers -- otherwise just reset.
     usage = usage ? usage.reset(localLookup) : Usage(self, y18n)
     validation = validation ? validation.reset(localLookup) : Validation(self, usage, y18n)
-    command = command ? command.reset() : Command(self, usage, validation)
+    command = command ? command.reset() : Command(self, usage, validation, globalMiddleware)
     if (!completion) completion = Completion(self, usage, command)
 
     completionCommand = null
@@ -130,9 +156,10 @@ function Yargs (processArgs, cwd, parentRequire) {
   self.resetOptions()
 
   // temporary hack: allow "freezing" of reset-able state for parse(msg, cb)
-  let frozen
+  const frozens = []
   function freeze () {
-    frozen = {}
+    const frozen = {}
+    frozens.push(frozen)
     frozen.options = options
     frozen.configObjects = options.configObjects.slice(0)
     frozen.exitProcess = exitProcess
@@ -141,13 +168,18 @@ function Yargs (processArgs, cwd, parentRequire) {
     validation.freeze()
     command.freeze()
     frozen.strict = strict
+    frozen.strictCommands = strictCommands
     frozen.completionCommand = completionCommand
     frozen.output = output
     frozen.exitError = exitError
     frozen.hasOutput = hasOutput
     frozen.parsed = self.parsed
+    frozen.parseFn = parseFn
+    frozen.parseContext = parseContext
+    frozen.handlerFinishCommand = handlerFinishCommand
   }
   function unfreeze () {
+    const frozen = frozens.pop()
     options = frozen.options
     options.configObjects = frozen.configObjects
     exitProcess = frozen.exitProcess
@@ -160,10 +192,11 @@ function Yargs (processArgs, cwd, parentRequire) {
     validation.unfreeze()
     command.unfreeze()
     strict = frozen.strict
+    strictCommands = frozen.strictCommands
     completionCommand = frozen.completionCommand
-    parseFn = null
-    parseContext = null
-    frozen = undefined
+    parseFn = frozen.parseFn
+    parseContext = frozen.parseContext
+    handlerFinishCommand = frozen.handlerFinishCommand
   }
 
   self.boolean = function (keys) {
@@ -202,9 +235,18 @@ function Yargs (processArgs, cwd, parentRequire) {
     return self
   }
 
-  self.requiresArg = function (keys) {
-    argsert('<array|string>', [keys], arguments.length)
-    populateParserHintArray('requiresArg', keys)
+  self.requiresArg = function (keys, value) {
+    argsert('<array|string|object> [number]', [keys], arguments.length)
+    // If someone configures nargs at the same time as requiresArg,
+    // nargs should take precedent,
+    // see: https://github.com/yargs/yargs/pull/1572
+    // TODO: make this work with aliases, using a check similar to
+    // checkAllAliases() in yargs-parser.
+    if (typeof keys === 'string' && options.narg[keys]) {
+      return self
+    } else {
+      populateParserHintObject(self.requiresArg, false, 'narg', keys, NaN)
+    }
     return self
   }
 
@@ -312,7 +354,7 @@ function Yargs (processArgs, cwd, parentRequire) {
     argsert('[object|string] [string|function] [function]', [key, msg, parseFn], arguments.length)
     // allow a config object to be provided directly.
     if (typeof key === 'object') {
-      key = applyExtends(key, cwd)
+      key = applyExtends(key, cwd, self.getParserConfiguration()['deep-merge-config'])
       options.configObjects = (options.configObjects || []).concat(key)
       return self
     }
@@ -338,9 +380,9 @@ function Yargs (processArgs, cwd, parentRequire) {
     return self
   }
 
-  self.command = function (cmd, description, builder, handler) {
-    argsert('<string|array|object> [string|boolean] [function|object] [function]', [cmd, description, builder, handler], arguments.length)
-    command.addHandler(cmd, description, builder, handler)
+  self.command = function (cmd, description, builder, handler, middlewares) {
+    argsert('<string|array|object> [string|boolean] [function|object] [function] [array]', [cmd, description, builder, handler, middlewares], arguments.length)
+    command.addHandler(cmd, description, builder, handler, middlewares)
     return self
   }
 
@@ -416,6 +458,17 @@ function Yargs (processArgs, cwd, parentRequire) {
     return options.demandedCommands
   }
 
+  self.deprecateOption = function deprecateOption (option, message) {
+    argsert('<string> [string|boolean]', [option, message], arguments.length)
+    options.deprecatedOptions[option] = message
+    return self
+  }
+
+  self.getDeprecatedOptions = () => {
+    argsert([], 0)
+    return options.deprecatedOptions
+  }
+
   self.implies = function (key, value) {
     argsert('<string|object> [number|string|array]', [key, value], arguments.length)
     validation.implies(key, value)
@@ -457,6 +510,14 @@ function Yargs (processArgs, cwd, parentRequire) {
     return self
   }
 
+  self.onFinishCommand = function (f) {
+    argsert('<function>', [f], arguments.length)
+    handlerFinishCommand = f
+    return self
+  }
+
+  self.getHandlerFinishCommand = () => handlerFinishCommand
+
   self.check = function (f, _global) {
     argsert('<function> [boolean]', [f, _global], arguments.length)
     validation.check(f, _global !== false)
@@ -476,17 +537,17 @@ function Yargs (processArgs, cwd, parentRequire) {
     return self
   }
 
-  self.pkgConf = function pkgConf (key, path) {
-    argsert('<string> [string]', [key, path], arguments.length)
+  self.pkgConf = function pkgConf (key, rootPath) {
+    argsert('<string> [string]', [key, rootPath], arguments.length)
     let conf = null
     // prefer cwd to require-main-filename in this method
     // since we're looking for e.g. "nyc" config in nyc consumer
     // rather than "yargs" config in nyc (where nyc is the main filename)
-    const obj = pkgUp(path || cwd)
+    const obj = pkgUp(rootPath || cwd)
 
     // If an object exists in the key, add it to options.configObjects
     if (obj[key] && typeof obj[key] === 'object') {
-      conf = applyExtends(obj[key], path || cwd)
+      conf = applyExtends(obj[key], rootPath || cwd, self.getParserConfiguration()['deep-merge-config'])
       options.configObjects = (options.configObjects || []).concat(conf)
     }
 
@@ -494,16 +555,24 @@ function Yargs (processArgs, cwd, parentRequire) {
   }
 
   const pkgs = {}
-  function pkgUp (path) {
-    const npath = path || '*'
+  function pkgUp (rootPath) {
+    const npath = rootPath || '*'
     if (pkgs[npath]) return pkgs[npath]
     const findUp = require('find-up')
 
     let obj = {}
     try {
+      let startDir = rootPath || require('require-main-filename')(parentRequire || require)
+
+      // When called in an environment that lacks require.main.filename, such as a jest test runner,
+      // startDir is already process.cwd(), and should not be shortened.
+      // Whether or not it is _actually_ a directory (e.g., extensionless bin) is irrelevant, find-up handles it.
+      if (!rootPath && path.extname(startDir)) {
+        startDir = path.dirname(startDir)
+      }
+
       const pkgJsonPath = findUp.sync('package.json', {
-        cwd: path || require('require-main-filename')(parentRequire || require),
-        normalize: false
+        cwd: startDir
       })
       obj = JSON.parse(fs.readFileSync(pkgJsonPath))
     } catch (noop) {}
@@ -516,7 +585,15 @@ function Yargs (processArgs, cwd, parentRequire) {
   let parseContext = null
   self.parse = function parse (args, shortCircuit, _parseFn) {
     argsert('[string|array] [function|boolean|object] [function]', [args, shortCircuit, _parseFn], arguments.length)
-    if (typeof args === 'undefined') args = processArgs
+    freeze()
+    if (typeof args === 'undefined') {
+      const argv = self._parseArgs(processArgs)
+      const tmpParsed = self.parsed
+      unfreeze()
+      // TODO: remove this compatibility hack when we release yargs@15.x:
+      self.parsed = tmpParsed
+      return argv
+    }
 
     // a context object can optionally be provided, this allows
     // additional information to be passed to a command handler.
@@ -536,10 +613,10 @@ function Yargs (processArgs, cwd, parentRequire) {
     // skipping validation, etc.
     if (!shortCircuit) processArgs = args
 
-    freeze()
     if (parseFn) exitProcess = false
 
     const parsed = self._parseArgs(args, shortCircuit)
+    completion.setParsed(self.parsed)
     if (parseFn) parseFn(exitError, parsed, output)
     unfreeze()
 
@@ -565,9 +642,15 @@ function Yargs (processArgs, cwd, parentRequire) {
 
       if (opt.alias) self.alias(key, opt.alias)
 
+      const deprecate = opt.deprecate || opt.deprecated
+
+      if (deprecate) {
+        self.deprecateOption(key, deprecate)
+      }
+
       const demand = opt.demand || opt.required || opt.require
 
-      // deprecated, use 'demandOption' instead
+      // A required option can be specified via "demand: true".
       if (demand) {
         self.demand(key, demand)
       }
@@ -649,8 +732,9 @@ function Yargs (processArgs, cwd, parentRequire) {
       }
 
       const desc = opt.describe || opt.description || opt.desc
-      if (!opt.hidden) {
-        self.describe(key, desc)
+      self.describe(key, desc)
+      if (opt.hidden) {
+        self.hide(key)
       }
 
       if (opt.requiresArg) {
@@ -669,8 +753,8 @@ function Yargs (processArgs, cwd, parentRequire) {
     }
 
     // .positional() only supports a subset of the configuration
-    // options availble to .option().
-    const supportedOpts = ['default', 'implies', 'normalize',
+    // options available to .option().
+    const supportedOpts = ['default', 'defaultDescription', 'implies', 'normalize',
       'choices', 'conflicts', 'coerce', 'type', 'describe',
       'desc', 'description', 'alias']
     opts = objFilter(opts, (k, v) => {
@@ -740,6 +824,22 @@ function Yargs (processArgs, cwd, parentRequire) {
   }
   self.getStrict = () => strict
 
+  let strictCommands = false
+  self.strictCommands = function (enabled) {
+    argsert('[boolean]', [enabled], arguments.length)
+    strictCommands = enabled !== false
+    return self
+  }
+  self.getStrictCommands = () => strictCommands
+
+  let parserConfig = {}
+  self.parserConfiguration = function parserConfiguration (config) {
+    argsert('<object>', [config], arguments.length)
+    parserConfig = config
+    return self
+  }
+  self.getParserConfiguration = () => parserConfig
+
   self.showHelp = function (level) {
     argsert('[string|function]', [level], arguments.length)
     if (!self.parsed) self._parseArgs(processArgs) // run parser, if it has not already been executed.
@@ -816,6 +916,28 @@ function Yargs (processArgs, cwd, parentRequire) {
     return self
   }
 
+  const defaultShowHiddenOpt = 'show-hidden'
+  options.showHiddenOpt = defaultShowHiddenOpt
+  self.addShowHiddenOpt = self.showHidden = function addShowHiddenOpt (opt, msg) {
+    argsert('[string|boolean] [string]', [opt, msg], arguments.length)
+
+    if (arguments.length === 1) {
+      if (opt === false) return self
+    }
+
+    const showHiddenOpt = typeof opt === 'string' ? opt : defaultShowHiddenOpt
+    self.boolean(showHiddenOpt)
+    self.describe(showHiddenOpt, msg || usage.deferY18nLookup('Show hidden options'))
+    options.showHiddenOpt = showHiddenOpt
+    return self
+  }
+
+  self.hide = function hide (key) {
+    argsert('<string|object>', [key], arguments.length)
+    options.hiddenOptions.push(key)
+    return self
+  }
+
   self.showHelpOnFail = function showHelpOnFail (enabled, message) {
     argsert('[boolean|string] [string]', [enabled, message], arguments.length)
     usage.showHelpOnFail(enabled, message)
@@ -846,9 +968,9 @@ function Yargs (processArgs, cwd, parentRequire) {
     }
 
     // register the completion command.
-    completionCommand = cmd || 'completion'
+    completionCommand = cmd || completionCommand || 'completion'
     if (!desc && desc !== false) {
-      desc = 'generate bash completion script'
+      desc = 'generate completion script'
     }
     self.command(completionCommand, desc)
 
@@ -858,10 +980,10 @@ function Yargs (processArgs, cwd, parentRequire) {
     return self
   }
 
-  self.showCompletionScript = function ($0) {
-    argsert('[string]', [$0], arguments.length)
+  self.showCompletionScript = function ($0, cmd) {
+    argsert('[string] [string]', [$0, cmd], arguments.length)
     $0 = $0 || self.$0
-    _logger.log(completion.generateCompletionScript($0, completionCommand))
+    _logger.log(completion.generateCompletionScript($0, cmd || completionCommand || 'completion'))
     return self
   }
 
@@ -958,13 +1080,21 @@ function Yargs (processArgs, cwd, parentRequire) {
     enumerable: true
   })
 
-  self._parseArgs = function parseArgs (args, shortCircuit, _skipValidation, commandIndex) {
-    let skipValidation = !!_skipValidation
+  self._parseArgs = function parseArgs (args, shortCircuit, _calledFromCommand, commandIndex) {
+    let skipValidation = !!_calledFromCommand
     args = args || processArgs
 
     options.__ = y18n.__
-    options.configuration = pkgUp()['yargs'] || {}
-    const parsed = Parser.detailed(args, options)
+    options.configuration = self.getParserConfiguration()
+
+    const populateDoubleDash = !!options.configuration['populate--']
+    const config = Object.assign({}, options.configuration, {
+      'populate--': true
+    })
+    const parsed = Parser.detailed(args, Object.assign({}, options, {
+      configuration: config
+    }))
+
     let argv = parsed.argv
     if (parseContext) argv = Object.assign({}, argv, parseContext)
     const aliases = parsed.aliases
@@ -979,7 +1109,7 @@ function Yargs (processArgs, cwd, parentRequire) {
       // are two passes through the parser. If completion
       // is being performed short-circuit on the first pass.
       if (shortCircuit) {
-        return argv
+        return (populateDoubleDash || _calledFromCommand) ? argv : self._copyDoubleDash(argv)
       }
 
       // if there's a handler associated with a
@@ -997,8 +1127,11 @@ function Yargs (processArgs, cwd, parentRequire) {
           argv[helpOpt] = true
         }
       }
+
       const handlerKeys = command.getCommands()
-      const skipDefaultCommand = argv[helpOpt] && (handlerKeys.length > 1 || handlerKeys[0] !== '$0')
+      const requestCompletions = completion.completionKey in argv
+      const skipRecommendation = argv[helpOpt] || requestCompletions
+      const skipDefaultCommand = skipRecommendation && (handlerKeys.length > 1 || handlerKeys[0] !== '$0')
 
       if (argv._.length) {
         if (handlerKeys.length) {
@@ -1006,11 +1139,11 @@ function Yargs (processArgs, cwd, parentRequire) {
           for (let i = (commandIndex || 0), cmd; argv._[i] !== undefined; i++) {
             cmd = String(argv._[i])
             if (~handlerKeys.indexOf(cmd) && cmd !== completionCommand) {
-              setPlaceholderKeys(argv)
               // commands are executed using a recursive algorithm that executes
               // the deepest command first; we keep track of the position in the
               // argv._ array that is currently being executed.
-              return command.runCommand(cmd, self, parsed, i + 1)
+              const innerArgv = command.runCommand(cmd, self, parsed, i + 1)
+              return populateDoubleDash ? innerArgv : self._copyDoubleDash(innerArgv)
             } else if (!firstUnknownCommand && cmd !== completionCommand) {
               firstUnknownCommand = cmd
               break
@@ -1019,31 +1152,31 @@ function Yargs (processArgs, cwd, parentRequire) {
 
           // run the default command, if defined
           if (command.hasDefaultCommand() && !skipDefaultCommand) {
-            setPlaceholderKeys(argv)
-            return command.runCommand(null, self, parsed)
+            const innerArgv = command.runCommand(null, self, parsed)
+            return populateDoubleDash ? innerArgv : self._copyDoubleDash(innerArgv)
           }
 
           // recommend a command if recommendCommands() has
           // been enabled, and no commands were found to execute
-          if (recommendCommands && firstUnknownCommand && !argv[helpOpt]) {
+          if (recommendCommands && firstUnknownCommand && !skipRecommendation) {
             validation.recommendCommands(firstUnknownCommand, handlerKeys)
           }
         }
 
         // generate a completion script for adding to ~/.bashrc.
-        if (completionCommand && ~argv._.indexOf(completionCommand) && !argv[completion.completionKey]) {
+        if (completionCommand && ~argv._.indexOf(completionCommand) && !requestCompletions) {
           if (exitProcess) setBlocking(true)
           self.showCompletionScript()
           self.exit(0)
         }
       } else if (command.hasDefaultCommand() && !skipDefaultCommand) {
-        setPlaceholderKeys(argv)
-        return command.runCommand(null, self, parsed)
+        const innerArgv = command.runCommand(null, self, parsed)
+        return populateDoubleDash ? innerArgv : self._copyDoubleDash(innerArgv)
       }
 
       // we must run completions first, a user might
       // want to complete the --help or --version option.
-      if (completion.completionKey in argv) {
+      if (requestCompletions) {
         if (exitProcess) setBlocking(true)
 
         // we allow for asynchronous completions,
@@ -1056,7 +1189,7 @@ function Yargs (processArgs, cwd, parentRequire) {
 
           self.exit(0)
         })
-        return setPlaceholderKeys(argv)
+        return (populateDoubleDash || _calledFromCommand) ? argv : self._copyDoubleDash(argv)
       }
 
       // Handle 'help' and 'version' options
@@ -1091,7 +1224,7 @@ function Yargs (processArgs, cwd, parentRequire) {
 
         // if we're executed via bash completion, don't
         // bother with validation.
-        if (!argv[completion.completionKey]) {
+        if (!requestCompletions) {
           self._runValidation(argv, aliases, {}, parsed.error)
         }
       }
@@ -1100,15 +1233,36 @@ function Yargs (processArgs, cwd, parentRequire) {
       else throw err
     }
 
-    return setPlaceholderKeys(argv)
+    return (populateDoubleDash || _calledFromCommand) ? argv : self._copyDoubleDash(argv)
+  }
+
+  // to simplify the parsing of positionals in commands,
+  // we temporarily populate '--' rather than _, with arguments
+  // after the '--' directive. After the parse, we copy these back.
+  self._copyDoubleDash = function (argv) {
+    if (!argv._ || !argv['--']) return argv
+    argv._.push.apply(argv._, argv['--'])
+
+    // TODO(bcoe): refactor command parsing such that this delete is not
+    // necessary: https://github.com/yargs/yargs/issues/1482
+    try {
+      delete argv['--']
+    } catch (_err) {}
+
+    return argv
   }
 
   self._runValidation = function runValidation (argv, aliases, positionalMap, parseErrors) {
     if (parseErrors) throw new YError(parseErrors.message)
     validation.nonOptionCount(argv)
-    validation.missingArgumentValue(argv)
     validation.requiredArguments(argv)
-    if (strict) validation.unknownArguments(argv, aliases, positionalMap)
+    let failedStrictCommands = false
+    if (strictCommands) {
+      failedStrictCommands = validation.unknownCommands(argv, aliases, positionalMap)
+    }
+    if (strict && !failedStrictCommands) {
+      validation.unknownArguments(argv, aliases, positionalMap)
+    }
     validation.customChecks(argv, aliases)
     validation.limitedChoices(argv)
     validation.implications(argv)
@@ -1117,24 +1271,8 @@ function Yargs (processArgs, cwd, parentRequire) {
 
   function guessLocale () {
     if (!detectLocale) return
-
-    try {
-      const osLocale = require('os-locale')
-      self.locale(osLocale.sync({ spawn: false }))
-    } catch (err) {
-      // if we explode looking up locale just noop
-      // we'll keep using the default language 'en'.
-    }
-  }
-
-  function setPlaceholderKeys (argv) {
-    Object.keys(options.key).forEach((key) => {
-      // don't set placeholder keys for dot
-      // notation options 'foo.bar'.
-      if (~key.indexOf('.')) return
-      if (typeof argv[key] === 'undefined') argv[key] = undefined
-    })
-    return argv
+    const locale = process.env.LC_ALL || process.env.LC_MESSAGES || process.env.LANG || process.env.LANGUAGE || 'en_US'
+    self.locale(locale.replace(/[.:].*/, ''))
   }
 
   // an app should almost always have --version and --help,
@@ -1145,6 +1283,9 @@ function Yargs (processArgs, cwd, parentRequire) {
   return self
 }
 
+// allow consumers to directly use the version of yargs-parser used by yargs
+exports.Parser = Parser
+
 // rebase an absolute path to a relative one with respect to a base directory
 // exported for tests
 exports.rebase = rebase
